/*
        Resource.c 1.1 for X11R6 & GNU C Compiler

        Copyright (C) 1998 Syllabub
        Maroontress Fast Software.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "Resource.h"

int
FQLength(char *base, char *componet)
{
    return (strlen(base) + 1 + strlen(componet));
}

void
FQCompose(char *base, char *componet, char *buf)
{
    (void)sprintf(buf, "%s.%s", base, componet);
}

void
GetPreference(XrmDatabase db, char *name, char *class, SylSetting *s)
{
    char *fq_name, *fq_class, *type;
    XrmValue value;

    s->spec = NULL;
    if ((fq_name = (char *)alloca(FQLength(name, s->name) + 1)) == NULL)
	return;
    FQCompose(name, s->name, fq_name);
    if ((fq_class = (char *)alloca(FQLength(class, s->class) + 1)) == NULL)
	return;
    FQCompose(class, s->class, fq_class);
#if 0
    fprintf(stderr, "%s, %s\n", fq_name, fq_class);
#endif
#if 0
    printf("%s: %s\n", fq_name, s->orig);
    printf("%s: %s\n", fq_class, s->orig);
#endif
    if (XrmGetResource(db, fq_name, fq_class, &type, &value) == False)
	return;
    s->spec = (char *)value.addr;
}

void
LoadColorset(Display *disp, char *base_name, SylSetting **s, unsigned long *p)
{
    Colormap cmap;
    XColor color;
    char *cname;

    cmap = DefaultColormap(disp, DefaultScreen(disp));
    for (; *s != NULL; ++p, ++s) {
	if ((cname = (*s)->spec) == NULL)
	    cname = (*s)->orig;
        if (XParseColor(disp, cmap, cname, &color) == 0) {
            fprintf(stderr, "%s.%s: color \"%s\" is not available.\n",
		    base_name, (*s)->name, cname);
            continue;
        }
        if (XAllocColor(disp, cmap, &color) == 0) {
            fprintf(stderr, "%s.%s: no available color cell for \"%s\".\n",
		    base_name, (*s)->name, cname);
            continue;
        }
        *p = color.pixel;
    }
}

static XFontSet
CreateFontset(Display *disp, char *fontlist, SylFontSet *fsp)
{
    XFontSet fs;
    int n_missings, n_fontinfos;
    char **missing, *alternation, **fontname;
    XFontStruct **fontinfo;

    fs = XCreateFontSet(disp, fontlist, &missing, &n_missings, &alternation);
    if (n_missings > 0) {
        XFreeStringList(missing);
	return (None);
    }
    if ((n_fontinfos = XFontsOfFontSet(fs, &fontinfo, &fontname)) < 1) {
	XFreeFontSet(disp, fs);
	return (None);
    }
    fsp->id = fs;
    fsp->width = abs(XmbTextEscapement(fs, "x", 1));
    fsp->height = fontinfo[0]->ascent + fontinfo[0]->descent;
    fsp->ascent = fontinfo[0]->ascent;
    fsp->descent = fontinfo[0]->descent;
    return (fs);
}

void
LoadFontset(Display *disp, char *base_name, SylSetting **s, SylFontSet *p)
{
    char *fname;

    for (; *s != NULL; ++p, ++s) {
        if ((fname = (*s)->spec) == NULL)
            fname = (*s)->orig;
        if (CreateFontset(disp, fname, p) == None) {
            fprintf(stderr, "%s.%s: fontset \"%s\" is not available.\n",
                    base_name, (*s)->name, fname);
        }
    }
}

void
SetFQClassHint(Display *disp, Window win, char *name, char *class,
	       Window target)
{
    char *fq_name, *fq_class;
    XClassHint ch, fq;

    if (XGetClassHint(disp, win, &ch) == 0)
	return;
    if ((fq_name = (char *)alloca(FQLength(ch.res_name, name) + 1)) == NULL)
	goto error_short_memory; 
    FQCompose(ch.res_name, name, fq_name);
    if ((fq_class = (char *)alloca(FQLength(ch.res_class, class) + 1)) == NULL)
	goto error_short_memory; 
    FQCompose(ch.res_class, class, fq_class);
    fq.res_name = fq_name;
    fq.res_class = fq_class;
    XSetClassHint(disp, target, &fq);
 error_short_memory:
    XFree(ch.res_name);
    XFree(ch.res_class);
}

#define STRING_PARSE_EMPTYLINES       0
#define STRING_PARSE_KEYWORD          1
#define	STRING_PARSE_PUNCTUATOR       2
#define STRING_PARSE_NUMBER           3
#define STRING_PARSE_OPERATOR         4
#define STRING_PARSE_TOO_LONG_KEYWORD 5
#define STRING_PARSE_UNKNOWN_OPERATOR 6

typedef struct {
    char *ptr;
} StringParse;

static int
GetChar(StringParse *p)
{
    int c;

    if ((c = *(p->ptr)) == 0)
	return (EOF);
    ++(p->ptr);
    return (c);
}

static void
UngetChar(int c, StringParse *p)
{
    if (c != EOF)
	--(p->ptr);
}

static int
IsDelimiter(int c)
{
    return (strchr(" \t\n", c) != NULL);
}

static int
IsName(int c)
{
    return (isalnum(c) || c == '_');
}
#define IsFirstName(c) IsName(c)

static int
IsPuncuator(int c)
{
    return (strchr("[](){},;:", c) != NULL);
}

static int
GetOperator(StringParse *p, char *buf, int c)
{
    int w;

    *buf++ = c;
    if (c == '-') { /* Group1: "X" "X>" "XX" "X=" */
	if ((w = GetChar(p)) == '>' || w == '-' || w == '=')
	    *buf++ = w;
	else
	    UngetChar(w, p);
    }
    else if (strchr("+&|=", c) != NULL) { /* Group2: "X" "XX" "X=" */
	if ((w = GetChar(p)) == c || w == '=')
	    *buf++ = w;
	else
	    UngetChar(w, p);
    }
    else if (c == '<' || c == '>') { /* Group3: "X" "XX" "X=" "XX=" */
	if ((w = GetChar(p)) == c)
	    *buf++ = w;
	else
	    UngetChar(w, p);
	if ((w = GetChar(p)) == '=')
	    *buf++ = w;
	else
	    UngetChar(w, p);
    }
    else if (strchr("*%^!", c) != NULL) { /* Group4: "X" "X=" */
	if ((w = GetChar(p)) == '=')
	    *buf++ = w;
	else
	    UngetChar(w, p);
    }
    else if (c == '.') { /* Group5: "X" "XXX" */
	if ((w = GetChar(p)) != c)
	    UngetChar(w, p);	
	else if ((w = GetChar(p)) != (*buf++ = c)) {
	    *buf = 0;
	    return (1); /* parse error */
	}
	else
	    *buf++ = c;
    }
    else if (c == '/') { /* Group5: "X" "X=" "X*" */
	if ((w = GetChar(p)) == '*' || w == '=')
	    *buf++ = w;
	else
	    UngetChar(w, p);
    }
    else if (strchr("#?\"\'", c) == NULL) {
	*buf = 0;
	return (1); /* parse error */
    }
    *buf = 0;
    return (0);
}

static int
GetTokenFromString(StringParse *p, char *top)
{
    int c, n = 0;
    char *buf = top;

    while ((c = GetChar(p)) != EOF && IsDelimiter(c))
	;
    if (c == EOF) {
	*buf = 0;
	return (EOF);
    }
    else if (IsFirstName(c)) {
	for (n = 0, *buf++ = c; IsName(c = GetChar(p)) && n < 64; ++buf, ++n)
	    *buf = c;
	UngetChar(c, p);
	*buf = 0;
	if (n == 64 && IsName(c))
	    return (STRING_PARSE_TOO_LONG_KEYWORD);
	return (STRING_PARSE_KEYWORD);
    }
    else if (IsPuncuator(c)) {
	buf[0] = c;
	buf[1] = 0;
	return (STRING_PARSE_PUNCTUATOR);
    }
    else if (isdigit(c)) {
	*buf++ = c;
	if (c != '0') { /* 10$B?JDj?t(B */
	    while (isdigit(c = GetChar(p)))
		*buf++ = c;
	}
	else if (tolower(c = GetChar(p)) == 'x') { /* 16$B?JDj?t(B */
	    for (*buf++ = c; isxdigit(c = GetChar(p)); ++buf)
		*buf = c;
	}	    
	else if (isdigit(c)) { /* 8$B?JDj?t(B */
	    for (*buf++ = c; isdigit(c = GetChar(p)); ++buf)
		*buf = c;
	}
	/* $B$=$&$G$J$1$l$P(B0 */
	while (tolower(c) == 'u' || tolower(c) == 'l') { /* $B@\Hx;R(B */
	    *buf++ = c;
	    c = GetChar(p);
	}
	UngetChar(c, p);
	*buf = 0;
	return (STRING_PARSE_NUMBER);
    }
    else if (GetOperator(p, buf, c) == 0)
	return (STRING_PARSE_OPERATOR);
    return (STRING_PARSE_UNKNOWN_OPERATOR);
}

static int
GetNextWord(StringParse *p, char *w)
{
    int type;

    while ((type = GetTokenFromString(p, w)) == STRING_PARSE_EMPTYLINES)
	;
    return (type);
}

static int
AddKeymap(SylKeymap **map, unsigned int mod, unsigned int func)
{
    SylKeymap *cur;

    if ((cur = (SylKeymap *)malloc(sizeof(SylKeymap))) == NULL)
	return (1);
    cur->mod = mod;
    cur->func = func;
    cur->next = *map;
    *map = cur;
    return (0);
}

typedef struct {
    char *name;
    unsigned int code;
} ModifierTable;
static ModifierTable Modifier[] = {
    {"Shift",   ShiftMask},
    {"Control", ControlMask},
    {"Meta",    Mod1Mask}};
#define N_MODIFIERS (sizeof(Modifier) / sizeof(ModifierTable))

static int
GetModifier(char *name, int *mod)
{
    unsigned int n;

    for (n = 0; n < N_MODIFIERS; ++n) {
	if (strcmp(Modifier[n].name, name) == 0) {
	    *mod |= Modifier[n].code;
	    return (1);
	}
    }
    return (0);
}

static int
GetKeybinding(Display *disp, char *base, char *name, int func, char *s,
	      SylKeymap **map)
{
    int type;
    unsigned int mod;
    KeyCode code;
    KeySym sym;
    StringParse p;
    char w[64];

    p.ptr = s;
    do {
	mod = 0;
	while ((type = GetNextWord(&p, w)) == STRING_PARSE_KEYWORD
	       && GetModifier(w, &mod)) {
	    if ((type = GetNextWord(&p, w)) == EOF) {
		fprintf(stderr, "%s.%s: unexpected EOS.\n", base, name);
		return (1);
	    }
	    else if (strcmp(w, "-") != 0) {
		fprintf(stderr, "%s.%s: %s, parse error.\n", base, name, w);
		return (1);
	    }
	}
	if (type == EOF) {
	    fprintf(stderr, "%s.%s: unexpected EOS.\n", base, name);
	    return (1);
	}
	else if (type != STRING_PARSE_KEYWORD) {
	    fprintf(stderr, "%s.%s: %s, parse error.\n", base, name, w);
	    return (1);
	}
	else if ((sym = XStringToKeysym(w)) == NoSymbol) {
	    fprintf(stderr, "%s.%s: %s, unknown KeySym.\n", base, name, w);
	    return (1);
	}
	else if ((code = XKeysymToKeycode(disp, sym)) == 0) {
	    fprintf(stderr, "%s.%s: %s, unknown KeyCode.\n", base, name, w);
	    return (1);
	}
	else if (AddKeymap(&map[code], mod, func)) {
	    fprintf(stderr, "%s.%s: too short memory.\n", base, name);
	    return (1);
	}
    } while ((type = GetNextWord(&p, w)) == STRING_PARSE_PUNCTUATOR
	     && strcmp(w, ",") == 0);
    if (type != EOF) {
	fprintf(stderr, "%s.%s: %s, parse error.\n", base, name, w);
	return (1);
    }
    return (0);
}

void
LoadKeybinding(Display *disp, char *base, SylSetting *set, int n_sets,
	       SylKeymap **map)
{
    int n;
    char *str;

    for (n = 0; n < n_sets; ++n) {
	if ((str = set[n].spec) == NULL)
	    str = set[n].orig;
	GetKeybinding(disp, base, set[n].name, n, str, map);
    }
}
